from slang.stypes import Waveforms, Chunker, Featurizer, Snipper


def just_give_me_any_snipper() -> Snipper:
    pass


def fit_chunker_from_wf_gen(chunker: Chunker, wf_gen: Waveforms) -> Chunker:
    pass


def fit_featuriizer_from_wf_gen(
    featurizer: Featurizer, wf_gen: Waveforms
) -> Featurizer:
    pass
